@@ -20,11 +20,16 @@ limitations under the License.
 
 #define EIGEN_USE_THREADS
 
+#include <algorithm>
+#include <cstdint>
+#include <functional>
 #include <type_traits>
 #include <utility>
 #include <vector>
 
-#include "third_party/eigen3/unsupported/Eigen/CXX11/Tensor"
+#include "Eigen/Core"  // from @eigen_archive
+#include "unsupported/Eigen/CXX11/Tensor"  // from @eigen_archive
+#include "tensorflow/core/framework/bfloat16.h"
 #include "tensorflow/core/framework/op_kernel.h"
 #include "tensorflow/core/framework/register_types.h"
 #include "tensorflow/core/framework/tensor.h"
@@ -34,6 +39,7 @@ limitations under the License.
 #include "tensorflow/core/kernels/fill_functor.h"
 #include "tensorflow/core/lib/core/errors.h"
 #include "tensorflow/core/lib/gtl/inlined_vector.h"
+#include "tensorflow/core/platform/bfloat16.h"
 #include "tensorflow/core/platform/logging.h"
 #include "tensorflow/core/platform/types.h"
 #include "tensorflow/core/util/matmul_autotune.h"
@@ -41,19 +47,26 @@ limitations under the License.
 #include "tensorflow/core/util/work_sharder.h"
 
 #if defined(TENSORFLOW_USE_CUSTOM_CONTRACTION_KERNEL)
-#include "tensorflow/tsl/framework/contraction/eigen_contraction_kernel.h"
+#include "tsl/framework/contraction/eigen_contraction_kernel.h"
 #endif
 
 #if GOOGLE_CUDA || TENSORFLOW_USE_ROCM
+#include "xla/stream_executor/host_or_device_scalar.h"
 #include "tensorflow/core/kernels/gpu_utils.h"
+#include "tensorflow/core/kernels/matmul_util.h"
+#include "tensorflow/core/kernels/numeric_options_utils.h"
 #include "tensorflow/core/platform/stream_executor.h"
 #endif  // GOOGLE_CUDA || TENSORFLOW_USE_ROCM
 #if GOOGLE_CUDA
 #include "third_party/gpus/cuda/include/cuda.h"
-#include "tensorflow/compiler/xla/stream_executor/cuda/cuda_blas_lt.h"
-#include "tensorflow/compiler/xla/stream_executor/host_or_device_scalar.h"
-#include "tensorflow/core/kernels/matmul_util.h"
+#include "xla/stream_executor/cuda/cuda_blas_lt.h"
 #endif  // GOOGLE_CUDA
+#if TENSORFLOW_USE_ROCM
+#include "rocm/rocm_config.h"
+#if TF_HIPBLASLT
+#include "xla/stream_executor/rocm/hip_blas_lt.h"
+#endif
+#endif
 
 namespace tensorflow {
 
@@ -85,7 +98,7 @@ struct ParallelMatMulKernel {
   }
 
   static void Run(const OpKernelContext* context, const Tensor& in_x,
-                  const Tensor in_y, bool adj_x, bool adj_y, bool trans_x,
+                  const Tensor& in_y, bool adj_x, bool adj_y, bool trans_x,
                   bool trans_y, const MatMulBCast& bcast, Tensor* out,
                   int batch_size) {
     static_assert(IsComplex, "Complex type expected.");
@@ -162,6 +175,21 @@ struct ParallelMatMulKernel<Scalar, false> {
   }
 };
 
+// Basic y-combinator implementation.
+template <class Func>
+struct YCombinatorImpl {
+  Func func;
+  template <class... Args>
+  decltype(auto) operator()(Args&&... args) const {
+    return func(*this, std::forward<Args>(args)...);
+  }
+};
+
+template <class Func>
+YCombinatorImpl<std::decay_t<Func>> YCombinator(Func&& func) {
+  return YCombinatorImpl<std::decay_t<Func>>{std::forward<Func>(func)};
+}
+
 // Sequential batch matmul kernel that calls the regular Eigen matmul.
 // We prefer this over the tensor contraction because it performs
 // better on vector-matrix and matrix-vector products.
@@ -227,6 +255,156 @@ struct SequentialMatMulKernel {
     }
   }
 };
+
+// For single-batch multiplications, manually parallize by splitting the output
+// matrix.
+template <typename Scalar>
+struct SingleBatchParallelMatMulKernel {
+  using Matrix =
+      Eigen::Matrix<Scalar, Eigen::Dynamic, Eigen::Dynamic, Eigen::RowMajor>;
+  using ConstMatrixMap = Eigen::Map<const Matrix>;
+  using MatrixMap = Eigen::Map<Matrix>;
+
+  static ConstMatrixMap ConstTensorToEigenMatrix(const Tensor& t) {
+    return ConstMatrixMap(t.flat<Scalar>().data(), t.dim_size(1),
+                          t.dim_size(2));
+  }
+
+  static MatrixMap TensorToEigenMatrix(Tensor* t) {
+    return MatrixMap(t->flat<Scalar>().data(), t->dim_size(1), t->dim_size(2));
+  }
+
+  static void Run(const CPUDevice& device, const Tensor& in_x,
+                  const Tensor& in_y, bool adj_x, bool adj_y, bool trans_x,
+                  bool trans_y, Tensor* out) {
+    using Eigen::Index;
+    Eigen::ThreadPoolInterface* pool = device.getPool();
+
+    Index m = (trans_x || adj_x) ? in_x.dim_size(2) : in_x.dim_size(1);
+    Index k = (trans_x || adj_x) ? in_x.dim_size(1) : in_x.dim_size(2);
+    Index n = (trans_y || adj_y) ? in_y.dim_size(1) : in_y.dim_size(2);
+
+    auto x_mat = ConstTensorToEigenMatrix(in_x);
+    auto y_mat = ConstTensorToEigenMatrix(in_y);
+    auto out_mat = TensorToEigenMatrix(out);
+
+    // Computes a block of the output matrix.
+    auto compute_matmul_block = [&x_mat, &y_mat, &out_mat, adj_x, trans_x,
+                                 adj_y, trans_y](Index row, Index col,
+                                                 Index nrows, Index ncols) {
+      auto z = out_mat.block(row, col, nrows, ncols);
+
+      // Assume at most one of adj_x or trans_x is true. Similarly, for adj_y
+      // and trans_y.
+      if (!adj_x && !trans_x) {
+        auto x = x_mat.middleRows(row, nrows);
+        if (!adj_y && !trans_y) {
+          auto y = y_mat.middleCols(col, ncols);
+          z = x * y;
+        } else if (adj_y) {
+          auto y = y_mat.middleRows(col, ncols);
+          z.noalias() = x * y.adjoint();
+        } else {  // trans_y == true
+          auto y = y_mat.middleRows(col, ncols);
+          z.noalias() = x * y.transpose();
+        }
+      } else if (adj_x) {
+        auto x = x_mat.middleCols(row, nrows);
+        if (!adj_y && !trans_y) {
+          auto y = y_mat.middleCols(col, ncols);
+          z.noalias() = x.adjoint() * y;
+        } else if (adj_y) {
+          auto y = y_mat.middleRows(col, ncols);
+          z.noalias() = x.adjoint() * y.adjoint();
+        } else {  // trans_y == true
+          auto y = y_mat.middleRows(col, ncols);
+          z.noalias() = x.adjoint() * y.transpose();
+        }
+      } else {  // trans_x == true
+        auto x = x_mat.middleCols(row, nrows);
+        if (!adj_y && !trans_y) {
+          auto y = y_mat.middleCols(col, ncols);
+          z.noalias() = x.transpose() * y;
+        } else if (adj_y) {
+          auto y = y_mat.middleRows(col, ncols);
+          z.noalias() = x.transpose() * y.adjoint();
+        } else {  // trans_y == true
+          auto y = y_mat.middleRows(col, ncols);
+          z.noalias() = x.transpose() * y.transpose();
+        }
+      }
+    };
+
+    // Split the work across n threads, unless the total amount of work
+    // is small (e.g. 128 * 128) - in which case use fewer threads.  This is
+    // the same heuristic value used in LaunchBatchMatMul below.
+    const int64_t kMaxCostOuterParallelism = 128 * 128;
+    Index work_limit = std::max<Index>((m * k * n) / pool->NumThreads(),
+                                       kMaxCostOuterParallelism);
+    // Blocks should have a size no smaller than 8 * kPacketSize, except perhaps
+    // for tail blocks.
+    constexpr int kPacketSize = Eigen::internal::packet_traits<Scalar>::size;
+    constexpr Index kBlockMin = 8 * kPacketSize;
+
+    // Precompute how many blocks there will be.
+    auto compute_blocks = YCombinator([k, work_limit, kBlockMin](
+                                          auto& compute_blocks, Index row,
+                                          Index col, Index nrows,
+                                          Index ncols) -> Index {
+      Index work = nrows * k * ncols;
+      Index blocks = 0;
+      while (work > work_limit && (nrows > kBlockMin || ncols > kBlockMin)) {
+        if (nrows > ncols) {
+          Index half = Eigen::divup(nrows / 2, kBlockMin) * kBlockMin;
+          blocks += 1 + compute_blocks(row + half, col, nrows - half, ncols);
+          nrows = half;
+        } else {
+          Index half = Eigen::divup(ncols / 2, kBlockMin) * kBlockMin;
+          blocks += 1 + compute_blocks(row, col + half, nrows, ncols - half);
+          ncols = half;
+        }
+        work = nrows * k * ncols;
+      }
+      return blocks;
+    });
+    Index total_blocks = 1 + compute_blocks(0, 0, m, n);
+
+    // Recursively split work according to the exact same heuristic as above.
+    Eigen::Barrier barrier(total_blocks);
+    auto handle_range = YCombinator(
+        [k, pool, &barrier, work_limit, kBlockMin, &compute_matmul_block](
+            auto& handle_range, Index row, Index col, Index nrows,
+            Index ncols) -> void {
+          Index work = nrows * k * ncols;
+          while (work > work_limit &&
+                 (nrows > kBlockMin || ncols > kBlockMin)) {
+            if (nrows > ncols) {
+              Index half = Eigen::divup(nrows / 2, kBlockMin) * kBlockMin;
+              pool->Schedule([&handle_range, row, half, col, nrows, ncols]() {
+                handle_range(row + half, col, nrows - half, ncols);
+              });
+              nrows = half;
+            } else {
+              Index half = Eigen::divup(ncols / 2, kBlockMin) * kBlockMin;
+              pool->Schedule([&handle_range, row, half, col, nrows, ncols]() {
+                handle_range(row, col + half, nrows, ncols - half);
+              });
+              ncols = half;
+            }
+            work = nrows * k * ncols;
+          }
+
+          if (nrows > 0 && ncols > 0) {
+            // Compute the output block.
+            compute_matmul_block(row, col, nrows, ncols);
+          }
+          barrier.Notify();
+        });
+    handle_range(0, 0, m, n);
+    barrier.Wait();
+  }
+};
+
 }  // namespace
 
 template <typename Device, typename Scalar>
@@ -236,7 +414,8 @@ template <typename Scalar>
 struct LaunchBatchMatMul<CPUDevice, Scalar> {
   static void Launch(OpKernelContext* context, const Tensor& in_x,
                      const Tensor& in_y, bool adj_x, bool adj_y, bool trans_x,
-                     bool trans_y, const MatMulBCast& bcast, Tensor* out) {
+                     bool trans_y, bool grad_x, bool grad_y,
+                     const MatMulBCast& bcast, Tensor* out) {
     typedef ParallelMatMulKernel<Scalar, Eigen::NumTraits<Scalar>::IsComplex>
         ParallelMatMulKernel;
     bool conjugate_result = false;
@@ -261,7 +440,7 @@ struct LaunchBatchMatMul<CPUDevice, Scalar> {
       ParallelMatMulKernel::Run(context, in_x, in_y, adj_x, adj_y, trans_x,
                                 trans_y, bcast, out, batch_size);
       conjugate_result = adj_x;
-    } else {
+    } else if (batch_size > 1) {
       // Parallelize over outer dims. For small matrices and large batches, it
       // is counter-productive to parallelize the inner matrix multiplies.
       Shard(worker_threads.num_threads, worker_threads.workers, batch_size,
@@ -272,7 +451,17 @@ struct LaunchBatchMatMul<CPUDevice, Scalar> {
                                                   trans_x, trans_y, bcast, out,
                                                   start, limit);
             });
+    } else if (cost_per_unit > kMaxCostOuterParallelism) {
+      // Split along output blocks.
+      SingleBatchParallelMatMulKernel<Scalar>::Run(context->eigen_cpu_device(),
+                                                   in_x, in_y, adj_x, adj_y,
+                                                   trans_x, trans_y, out);
+    } else {
+      // Single small multiplication.
+      SequentialMatMulKernel<Scalar>::Run(in_x, in_y, adj_x, adj_y, trans_x,
+                                          trans_y, bcast, out, 0, batch_size);
     }
+
     if (conjugate_result) {
       // We used one of the identities
       //   conj(a) * conj(b) = conj(a * b)
@@ -284,7 +473,7 @@ struct LaunchBatchMatMul<CPUDevice, Scalar> {
   }
 };
 
-#if GOOGLE_CUDA
+#if GOOGLE_CUDA || TF_HIPBLASLT
 
 namespace {
 // A dummy type to group matmul autotune results together.
@@ -355,7 +544,8 @@ template <typename Scalar>
 struct LaunchBatchMatMul<GPUDevice, Scalar> {
   static void Launch(OpKernelContext* context, const Tensor& in_x,
                      const Tensor& in_y, bool adj_x, bool adj_y, bool trans_x,
-                     bool trans_y, const MatMulBCast& bcast, Tensor* out) {
+                     bool trans_y, bool grad_x, bool grad_y,
+                     const MatMulBCast& bcast, Tensor* out) {
     se::blas::Transpose trans[] = {se::blas::Transpose::kNoTranspose,
                                    se::blas::Transpose::kTranspose,
                                    se::blas::Transpose::kConjugateTranspose};
@@ -398,9 +588,26 @@ struct LaunchBatchMatMul<GPUDevice, Scalar> {
                                     std::is_same_v<Scalar, Eigen::bfloat16>;
     using Coefficient = std::conditional_t<is_16bit_input, float, Scalar>;
 
-#if GOOGLE_CUDA
+    se::blas::CallContext call_context = se::blas::CallContext::kNone;
+    OP_REQUIRES(context, grad_x == false || grad_y == false,
+                errors::InvalidArgument(
+                    "At least 1 of grad_x and grad_y shall be false"));
+    if (grad_x) {
+      call_context = se::blas::CallContext::kBackpropInput1;
+    }
+    if (grad_y) {
+      call_context = se::blas::CallContext::kBackpropInput2;
+    }
+#if GOOGLE_CUDA || TF_HIPBLASLT
     static const bool use_autotune = MatmulAutotuneEnable();
-    if (EnableCublasLtGemm()) {
+    bool bCublasLtSupport = true;
+#if TF_HIPBLASLT
+    if (!std::is_same_v<Scalar, float>) bCublasLtSupport = false;
+    auto cap = stream->GetRocmComputeCapability();
+    // as of ROCm 5.5, hipblaslt only supports MI200.
+    if (cap.gcn_arch_name().substr(0, 6) != "gfx90a") bCublasLtSupport = false;
+#endif
+    if (EnableCublasLtGemm() && bCublasLtSupport) {
       static const int64_t max_scratch_size =
           GetWorkspaceLimit(1LL << 32);  // 4GB by default
 
@@ -428,43 +635,39 @@ struct LaunchBatchMatMul<GPUDevice, Scalar> {
 
         std::optional<int> max_algorithm_count;
         if (!use_autotune) max_algorithm_count = 1;
-
-        auto plan_and_algorithms_or =
-            GetPlanAndAlgorithms(stream, matmul_params, max_algorithm_count);
+        absl::Mutex* pmu = nullptr;
+        auto plan_and_algorithms_or = GetPlanAndAlgorithms(
+            stream, matmul_params, &pmu, max_algorithm_count);
         OP_REQUIRES_OK(context, plan_and_algorithms_or.status());
+        absl::MutexLock lock(pmu);
         const auto* plan_and_algorithms =
             std::move(plan_and_algorithms_or).value();
-        const auto& plan = plan_and_algorithms->plan;
-        const auto& algorithms = plan_and_algorithms->algorithms;
-
-        se::cuda::BlasLt* blas_lt = se::cuda::GetBlasLt(stream);
-        OP_REQUIRES(context, blas_lt != nullptr,
-                    errors::Internal("blaslt not supported"));
+        auto n_algorithms = plan_and_algorithms->algorithms.size();
 
         se::blas::AlgorithmConfig algorithm_config(se::blas::kNoAlgorithm);
         if (!use_autotune) {
           algorithm_config.set_algorithm(0);
         } else if (!AutoTuneBatchMatmul::GetInstance()->Find(
                        matmul_params, &algorithm_config)) {
-          VLOG(4) << "Autotuning BlasLtMatmul over " << algorithms.size()
+          VLOG(4) << "Autotuning BlasLtMatmul over " << n_algorithms
                   << " algorithms.";
           se::blas::ProfileResult best_result;
           se::blas::ProfileResult profile_result;
 
-          for (size_t i = 0; i != algorithms.size(); ++i) {
-            const auto& profile_algorithm = algorithms[i];
+          for (size_t i = 0; i != n_algorithms; ++i) {
             // Create a new scratch allocator with every autotuning run so that
             // scratch space is deallocated between runs.
             BlasScratchAllocator scratch_allocator(context, max_scratch_size);
             Status cublas_launch_status =
-                DoBlasLtMatmul(stream, plan, *a_ptrs[0], *b_ptrs[0], *c_ptrs[0],
-                               profile_algorithm, scratch_allocator,
+                DoBlasLtMatmul(stream, *plan_and_algorithms, *a_ptrs[0],
+                               *b_ptrs[0], *c_ptrs[0], i, scratch_allocator,
                                /*bias = */ {}, &profile_result);
 
             VLOG(4) << "  Autotune algorithm " << i
                     << " result: " << profile_result.elapsed_time_in_ms()
                     << " ms, valid=" << profile_result.is_valid()
-                    << ", workspace_size=" << profile_algorithm.workspace_size;
+                    << ", workspace_size="
+                    << plan_and_algorithms->algorithms[i].workspace_size;
 
             if (cublas_launch_status.ok() && profile_result.is_valid() &&
                 profile_result.elapsed_time_in_ms() <
@@ -485,10 +688,8 @@ struct LaunchBatchMatMul<GPUDevice, Scalar> {
                                                      algorithm_config);
         }
         se::blas::AlgorithmType algorithm_idx = algorithm_config.algorithm();
-        OP_REQUIRES(context,
-                    0 <= algorithm_idx && algorithm_idx < algorithms.size(),
+        OP_REQUIRES(context, 0 <= algorithm_idx && algorithm_idx < n_algorithms,
                     errors::Internal("Missing/invalid BatchMatmul algorithm"));
-        const auto& algorithm = algorithms[algorithm_idx];
         BlasScratchAllocator scratch_allocator(context, max_scratch_size);
         VLOG(4) << "Calling BlasLtMatMul: a.shape=(" << bcast.x_batch_size()
                 << ", " << in_x.dim_size(1) << ", " << in_x.dim_size(2)
@@ -499,8 +700,9 @@ struct LaunchBatchMatMul<GPUDevice, Scalar> {
                 << "adj_x = " << adj_x << "adj_y = " << adj_y;
 
         OP_REQUIRES_OK(
-            context, DoBlasLtMatmul(stream, plan, *a_ptrs[0], *b_ptrs[0],
-                                    *c_ptrs[0], algorithm, scratch_allocator));
+            context,
+            DoBlasLtMatmul(stream, *plan_and_algorithms, *a_ptrs[0], *b_ptrs[0],
+                           *c_ptrs[0], algorithm_idx, scratch_allocator));
       } else {  // requires mixed broadcasting
         const std::vector<int64_t>& a_batch_indices = bcast.x_batch_indices();
         const std::vector<int64_t>& b_batch_indices = bcast.y_batch_indices();
@@ -518,15 +720,15 @@ struct LaunchBatchMatMul<GPUDevice, Scalar> {
         }
 
         BlasScratchAllocator scratch_allocator(context, max_scratch_size);
-        bool blas_launch_status =
-            stream
-                ->ThenBlasGemmBatchedWithScratch(
-                    blas_transpose_b, blas_transpose_a, n, m, k,
-                    static_cast<Coefficient>(1.0), b_ptrs,
-                    adj_y || trans_y ? k : n, a_ptrs, adj_x || trans_x ? m : k,
-                    static_cast<Coefficient>(0.0), c_ptrs, n, batch_size,
-                    &scratch_allocator)
-                .ok();
+        auto blas = stream->parent()->AsBlas();
+        OP_REQUIRES(context, blas != nullptr,
+                    absl::InternalError("No blas support for stream"));
+        bool blas_launch_status = blas->DoBlasGemmBatched(
+            stream, blas_transpose_b, blas_transpose_a, n, m, k,
+            static_cast<Coefficient>(1.0), b_ptrs, adj_y || trans_y ? k : n,
+            a_ptrs, adj_x || trans_x ? m : k, static_cast<Coefficient>(0.0),
+            c_ptrs, n, batch_size, GetNumericOptions(), &scratch_allocator,
+            call_context);
         if (!blas_launch_status) {
           context->SetStatus(errors::Internal(
               "Blas xGEMMBatched launch failed: a.shape=",
@@ -583,6 +785,9 @@ struct LaunchBatchMatMul<GPUDevice, Scalar> {
       // C' = B' x A', where ' stands for transpose (not adjoint).
       // TODO(yangzihao): Choose the best of the three strategies using
       // autotune.
+      auto blas = stream->parent()->AsBlas();
+      OP_REQUIRES(context, blas != nullptr,
+                  absl::InternalError("No blas support for stream"));
       if (batch_size == 1) {
         // This is a regular matrix*matrix or matrix*vector multiply. Avoid the
         // overhead of the scratch allocator and the batch interface.
@@ -601,14 +806,11 @@ struct LaunchBatchMatMul<GPUDevice, Scalar> {
                 blas_transpose_a == se::blas::Transpose::kTranspose
                     ? se::blas::Transpose::kNoTranspose
                     : se::blas::Transpose::kTranspose;
-            bool blas_launch_status =
-                stream
-                    ->ThenBlasGemv(gemv_trans_a, adj_x || trans_x ? m : k,
-                                   adj_x || trans_x ? k : m,
-                                   static_cast<Coefficient>(1.0), *(a_ptrs[0]),
-                                   adj_x || trans_x ? m : k, *(b_ptrs[0]), 1,
-                                   static_cast<Coefficient>(0.0), c_ptrs[0], 1)
-                    .ok();
+            bool blas_launch_status = blas->DoBlasGemv(
+                stream, gemv_trans_a, adj_x || trans_x ? m : k,
+                adj_x || trans_x ? k : m, static_cast<Coefficient>(1.0),
+                *(a_ptrs[0]), adj_x || trans_x ? m : k, *(b_ptrs[0]), 1,
+                static_cast<Coefficient>(0.0), c_ptrs[0], 1);
             if (!blas_launch_status) {
               context->SetStatus(errors::Internal(
                   "Blas xGEMV launch failed : a.shape=",
@@ -619,32 +821,29 @@ struct LaunchBatchMatMul<GPUDevice, Scalar> {
           }
         }
 
-        OP_REQUIRES_OK(context,
-                       stream->ThenBlasGemm(
-                           blas_transpose_b, blas_transpose_a, n, m, k,
+        OP_REQUIRES_OK(
+            context,
+            blas->BlasGemm(stream, blas_transpose_b, blas_transpose_a, n, m, k,
                            *(b_ptrs[0]), adj_y || trans_y ? k : n, *(a_ptrs[0]),
                            adj_x || trans_x ? m : k, c_ptrs[0], n,
-                           se::blas::kDefaultComputePrecision));
+                           GetNumericOptions(), call_context));
       } else if (use_strided_batched) {
         OP_REQUIRES_OK(
-            context, stream->ThenBlasGemmStridedBatched(
-                         blas_transpose_b, blas_transpose_a, n, m, k,
+            context, blas->BlasGemmStridedBatched(
+                         stream, blas_transpose_b, blas_transpose_a, n, m, k,
                          static_cast<Coefficient>(1.0), *b_ptrs[0],
                          adj_y || trans_y ? k : n, b_stride, *a_ptrs[0],
                          adj_x || trans_x ? m : k, a_stride,
                          static_cast<Coefficient>(0.0), c_ptrs[0], n, c_stride,
-                         batch_size, se::blas::kDefaultComputePrecision));
+                         batch_size, GetNumericOptions(), call_context));
       } else {
         BlasScratchAllocator scratch_allocator(context);
-        bool blas_launch_status =
-            stream
-                ->ThenBlasGemmBatchedWithScratch(
-                    blas_transpose_b, blas_transpose_a, n, m, k,
-                    static_cast<Coefficient>(1.0), b_ptrs,
-                    adj_y || trans_y ? k : n, a_ptrs, adj_x || trans_x ? m : k,
-                    static_cast<Coefficient>(0.0), c_ptrs, n, batch_size,
-                    &scratch_allocator)
-                .ok();
+        bool blas_launch_status = blas->DoBlasGemmBatched(
+            stream, blas_transpose_b, blas_transpose_a, n, m, k,
+            static_cast<Coefficient>(1.0), b_ptrs, adj_y || trans_y ? k : n,
+            a_ptrs, adj_x || trans_x ? m : k, static_cast<Coefficient>(0.0),
+            c_ptrs, n, batch_size, GetNumericOptions(), &scratch_allocator,
+            call_context);
         if (!blas_launch_status) {
           context->SetStatus(errors::Internal(
               "Blas xGEMMBatched launch failed : a.shape=",
@@ -653,7 +852,7 @@ struct LaunchBatchMatMul<GPUDevice, Scalar> {
               ", k=", k, ", batch_size=", batch_size));
         }
       }
-#if GOOGLE_CUDA
+#if GOOGLE_CUDA || TF_HIPBLASLT
     }
 #endif  // GOOGLE_CUDA
   }
@@ -661,6 +860,32 @@ struct LaunchBatchMatMul<GPUDevice, Scalar> {
 
 #endif  // GOOGLE_CUDA || TENSORFLOW_USE_ROCM
 
+template <typename T>
+inline void FastConvertToFloat(const T* src, float* dst, int64_t size) {
+  Eigen::Map<const Eigen::Array<T, Eigen::Dynamic, 1>> src_eigen(src, size);
+  Eigen::Map<Eigen::ArrayXf> dst_eigen(dst, size);
+  dst_eigen = src_eigen.template cast<float>();
+}
+
+template <typename T>
+inline void FastConvertFromFloat(const float* src, T* dst, int64_t size) {
+  Eigen::Map<const Eigen::ArrayXf> src_eigen(src, size);
+  Eigen::Map<Eigen::Array<T, Eigen::Dynamic, 1>> dst_eigen(dst, size);
+  dst_eigen = src_eigen.template cast<T>();
+}
+
+template <>
+inline void FastConvertToFloat<bfloat16>(const bfloat16* src, float* dst,
+                                         int64_t size) {
+  BFloat16ToFloat(src, dst, size);
+}
+
+template <>
+inline void FastConvertFromFloat<bfloat16>(const float* src, bfloat16* dst,
+                                           int64_t size) {
+  FloatToBFloat16(src, dst, size);
+}
+
 template <typename Device, typename Ta, typename Tb, typename Tout>
 class BaseBatchMatMulOp : public OpKernel {
  public:
@@ -673,11 +898,15 @@ class BaseBatchMatMulOp : public OpKernel {
       OP_REQUIRES_OK(context, context->GetAttr("transpose_b", &trans_y_));
       adj_x_ = false;
       adj_y_ = false;
+      OP_REQUIRES_OK(context, context->GetAttr("grad_a", &grad_input_1_));
+      OP_REQUIRES_OK(context, context->GetAttr("grad_b", &grad_input_2_));
     } else {
       OP_REQUIRES_OK(context, context->GetAttr("adj_x", &adj_x_));
       OP_REQUIRES_OK(context, context->GetAttr("adj_y", &adj_y_));
       trans_x_ = false;
       trans_y_ = false;
+      OP_REQUIRES_OK(context, context->GetAttr("grad_x", &grad_input_1_));
+      OP_REQUIRES_OK(context, context->GetAttr("grad_y", &grad_input_2_));
     }
   }
 
@@ -742,8 +971,17 @@ class BaseBatchMatMulOp : public OpKernel {
                 out_reshaped.CopyFrom(*out, TensorShape({batch_size, d0, d3})),
                 errors::Internal("Failed to reshape output from ",
                                  out->shape().DebugString()));
-    if (std::is_same_v<Device, CPUDevice> && std::is_same_v<Ta, bfloat16> &&
-        std::is_same_v<Tb, bfloat16>) {
+
+    // b/307285203: There seems to be an overly aggressive compiler optimization
+    // that optimizes away these data pointers unless we explicitly check them.
+    OP_REQUIRES(ctx,
+                in0_reshaped.data() != nullptr &&
+                    in1_reshaped.data() != nullptr &&
+                    out_reshaped.data() != nullptr,
+                absl::InternalError("Null data pointer encountered."));
+    if constexpr (std::is_same_v<Device, CPUDevice> && std::is_same_v<Ta, Tb> &&
+                  (std::is_same_v<Ta, bfloat16> ||
+                   std::is_same_v<Ta, Eigen::half>)) {
       Tensor in0_reshaped_float, in1_reshaped_float, out_reshaped_float;
       OP_REQUIRES_OK(ctx, ctx->allocate_temp(DT_FLOAT, in0_reshaped.shape(),
                                              &in0_reshaped_float));
@@ -752,31 +990,32 @@ class BaseBatchMatMulOp : public OpKernel {
       OP_REQUIRES_OK(ctx, ctx->allocate_temp(DT_FLOAT, out_reshaped.shape(),
                                              &out_reshaped_float));
 
-      // TODO: Avoid extra copy to make bfloat16 matmul efficient on CPU.
-      BFloat16ToFloat(in0_reshaped.flat<bfloat16>().data(),
-                      in0_reshaped_float.flat<float>().data(),
-                      in0_reshaped.NumElements());
-      BFloat16ToFloat(in1_reshaped.flat<bfloat16>().data(),
-                      in1_reshaped_float.flat<float>().data(),
-                      in1_reshaped.NumElements());
+      // TODO: Avoid extra copy to make (b)float16 matmul efficient on CPU.
+      FastConvertToFloat(in0_reshaped.flat<Ta>().data(),
+                         in0_reshaped_float.flat<float>().data(),
+                         in0_reshaped.NumElements());
+      FastConvertToFloat(in1_reshaped.flat<Tb>().data(),
+                         in1_reshaped_float.flat<float>().data(),
+                         in1_reshaped.NumElements());
 
       LaunchBatchMatMul<Device, float>::Launch(
           ctx, in0_reshaped_float, in1_reshaped_float, adj_x_, adj_y_, trans_x_,
-          trans_y_, bcast, &out_reshaped_float);
-      FloatToBFloat16(out_reshaped_float.flat<float>().data(),
-                      out_reshaped.flat<bfloat16>().data(), out->NumElements());
+          trans_y_, grad_input_1_, grad_input_2_, bcast, &out_reshaped_float);
+      FastConvertFromFloat<Tout>(out_reshaped_float.flat<float>().data(),
+                                 out_reshaped.flat<Tout>().data(),
+                                 out->NumElements());
     } else {
       // Cast tensor to desired type to reuse Eigen.
       // TODO(b/178749687): remove this cast if Eigen supports this natively.
-      if (!std::is_same<Ta, Tout>::value) {
+      if constexpr (!std::is_same<Ta, Tout>::value) {
         in0_reshaped = CastTensor<Ta, Tout>(in0_reshaped);
       }
-      if (!std::is_same<Tb, Tout>::value) {
+      if constexpr (!std::is_same<Tb, Tout>::value) {
         in1_reshaped = CastTensor<Tb, Tout>(in1_reshaped);
       }
-      LaunchBatchMatMul<Device, Tout>::Launch(ctx, in0_reshaped, in1_reshaped,
-                                              adj_x_, adj_y_, trans_x_,
-                                              trans_y_, bcast, &out_reshaped);
+      LaunchBatchMatMul<Device, Tout>::Launch(
+          ctx, in0_reshaped, in1_reshaped, adj_x_, adj_y_, trans_x_, trans_y_,
+          grad_input_1_, grad_input_2_, bcast, &out_reshaped);
     }
   }
 
@@ -790,6 +1029,8 @@ class BaseBatchMatMulOp : public OpKernel {
   bool adj_y_ = false;
   bool trans_x_ = false;
   bool trans_y_ = false;
+  bool grad_input_1_ = false;
+  bool grad_input_2_ = false;
 
   // Cast `t` from `SrcT` to `DstT`.
   template <typename SrcT, typename DstT>
@@ -840,7 +1081,7 @@ class BatchMatMulOp : public BaseBatchMatMulOp<Device, Ta, Tb, Tout> {
         }
       }
     }
-    return OkStatus();
+    return absl::OkStatus();
   }
 };
 
@@ -866,7 +1107,7 @@ class BatchMatMulV2Op : public BaseBatchMatMulOp<Device, Ta, Tb, Tout> {
     if (in1.dims() < 2) {
       return errors::InvalidArgument("In[1] ndims must be >= 2: ", in1.dims());
     }
-    return OkStatus();
+    return absl::OkStatus();
   }
 };
 
